##############################################################################
# SCAN TEXT FOR BIBLE ADDRESSES
##############################################################################
#
# (c) 2003 www.SpiritAndTruth.org
# May be freely used without restriction.
# mailto:contact@SpiritAndTruth.org
#
#

import string
import re
import exceptions
import urllib

VERSION = '1.11'
HISTORY = '''

1.12    Fixed parsing error where "John 3:18-19,36" returned two address
        ranges: "John 3:18-19" and "John 3:36-19". The latter range should
        be "John 3:36".

        Fixed parsing error where floating addresses with explicit colon
        where incorrectly taken as verses (rather than chapters) in some
        situations.

        Fixed parsing error where multiple verse references in a single-chapter
        book without an explicit chapter caused duplicate output (e.g., "Jude
        12a, 14b;" resulted in "Jude 12a, 14b14b").

1.11    Improved parsing of floating chapter numbers.

1.10    More flexible address scanning. Added support for generating
        e-Sword address format.

1.9     Fixed -lbxpbb flag with verse ranges. Previously

            Gen. 1:1-5

        generated

            Gen. [[1:1-5|bible.1.1.1-5]]

        Now it generates

            Gen. [[1:1-5|bible.1.1.1-1.1.5]]

        Fixed a defect which caused lists of semicolon-separated
        chapters (without verses) to not get tagged.

1.8     Bible addresses which span a range of verses now include the
        entire range by default. (Previous versions represented the range
        by pointing to the first verse of the range.)

        Corrected a defect which mistook the abbreviation "Ac" for the book
        of Acts as possible subscripts on the preceding verse (e.g., Gen. 1:1a).

1.7     1. Added the option to ignore the alphabetical case of book names
            when searching for Bible addresses.

        2. Fixed a bug in logging of out-of-range chapters or verses.

        3. Added support for generating URLs to lookup addresses at www.StudyLight.org.

1.6     Fixed minor error in how verbose flag was used
        to pass file handle.

1.5     The verbose 'flag' can either be a boolean or a
        file handle in which case the file is used to log
        the verbose output (as well as being printed).

1.4     Improved information in "out of range" chapter and
        verse error messages.

1.3     Fixed defect where including the book name in
        the first chapter text led to subsequent chapters
        being tagged with the bookname.
        For instance, tagging
        
            John 3:16; 4:4

        Resulted in two tagged address strings:

            [John 3:16]; [John 3:16; 4:4]

1.2     Fixed address tagging problem where text has an
        explicit verse designator (e.g., 'v.4'), but no
        chapter has been seen in the near context.

1.1     Fixed subtle tagging problems with addresses
        of the form which are a list of chapters without
        colons or verses, e.g.,

            Revelation 7 and 9-13.
            Revelation 17, 18, and 19

1.0     Correctly tags assumed book, floating verses,
        ranges that are chapters without verses, connects 'and'
        to previous known verse for greater confidence of
        verses which follow 'and.'

'''

##############################################################################
# Bible address information.
##############################################################################

class AddressError( exceptions.Exception ) :
    def __init__(self, args=None ) :
        self.args = args

class Address :

    CHAPTER_DIVISION = ':'
    NUMBERS = '[0-9]+'
    ADDRESS_PATTERN = NUMBERS + CHAPTER_DIVISION + NUMBERS

    # Formal short names we use internally.
    BOOKNAMES = [

        'Gen.',
        'Ex.',
        'Lev.',
        'Num.', 
        'Deu.',
        'Jos.',
        'Jdg.',
        'Ru.', 
        '1S.',
        '2S.',
        '1K.',
        '2K.', 
        '1Chr.',
        '2Chr.',
        'Ezra',
        'Ne.', 
        'Est.',
        'Job',
        'Ps.',
        'Pr.', 
        'Ecc.',
        'Sos.',
        'Isa.',
        'Jer.', 
        'Lam.',
        'Eze.',
        'Dan.',
        'Hos.', 
        'Joel',
        'Amos',
        'Ob.',
        'Jonah', 
        'Mic.',
        'Nah.',
        'Hab.',
        'Zep.', 
        'Hag.',
        'Zec.',
        'Mal.',
        'Mat.', 
        'Mark',
        'Luke',
        'John',
        'Acts', 
        'Rom.',
        '1Cor.',
        '2Cor.',
        'Gal.', 
        'Eph.',
        'Php.',
        'Col.',
        '1Th.', 
        '2Th.',
        '1Ti.',
        '2Ti.',
        'Tit.', 
        'Phm.',
        'Heb.',
        'Jas.',
        '1Pe.', 
        '2Pe.',
        '1Jn.',
        '2Jn.',
        '3Jn.', 
        'Jude',
        'Rev.'

        ]

    # ESWORD BOOK NAMES (www.e-sword.net)
    BOOKNAMES_ESWORD = [

        'Gen',
        'Exo',
        'Lev',
        'Num', 
        'Deu',
        'Jos',
        'Jdg',
        'Rth', 
        '1Sa',
        '2Sa',
        '1Ki',
        '2Ki', 
        '1Ch',
        '2Ch',
        'Ezr',
        'Neh', 
        'Est',
        'Job',
        'Psa',
        'Pro', 
        'Ecc',
        'Son',
        'Isa',
        'Jer', 
        'Lam',
        'Eze',
        'Dan',
        'Hos', 
        'Joe',
        'Amo',
        'Oba',
        'Jon', 
        'Mic',
        'Nah',
        'Hab',
        'Zep', 
        'Hag',
        'Zec',
        'Mal',
        'Mat', 
        'Mar',
        'Luk',
        'Joh',
        'Act', 
        'Rom',
        '1Co',
        '2Co',
        'Gal', 
        'Eph',
        'Phi',
        'Col',
        '1Th', 
        '2Th',
        '1Ti',
        '2Ti',
        'Tit', 
        'Phm',
        'Heb',
        'Jam',
        '1Pe', 
        '2Pe',
        '1Jo',
        '2Jo',
        '3Jo', 
        'Jud',
        'Rev'

        ]

    # Map formal short names to Libronix book numbers.
    BOOKNAME_TO_LBX_NUMBER = {

        'Gen.'  : '1',
        'Ex.'   : '2',
        'Lev.'  : '3',
        'Num.'  : '4',
        'Deu.'  : '5',
        'Jos.'  : '6',
        'Jdg.'  : '7',
        'Ru.'   : '8',
        '1S.'   : '9',
        '2S.'   : '10',
        '1K.'   : '11',
        '2K.'   : '12',
        '1Chr.' : '13',
        '2Chr.' : '14',
        'Ezra'  : '15',
        'Ne.'   : '16',
        'Est.'  : '17',
        'Job'   : '18',
        'Ps.'   : '19',
        'Pr.'   : '20',
        'Ecc.'  : '21',
        'Sos.'  : '22',
        'Isa.'  : '23',
        'Jer.'  : '24',
        'Lam.'  : '25',
        'Eze.'  : '26',
        'Dan.'  : '27',
        'Hos.'  : '28',
        'Joel'  : '29',
        'Amos'  : '30',
        'Ob.'   : '31',
        'Jonah' : '32',
        'Mic.'  : '33',
        'Nah.'  : '34',
        'Hab.'  : '35',
        'Zep.'  : '36',
        'Hag.'  : '37',
        'Zec.'  : '38',
        'Mal.'  : '39',

        'Mat.'  : '61',
        'Mark'  : '62',
        'Luke'  : '63',
        'John'  : '64',
        'Acts'  : '65',
        'Rom.'  : '66',
        '1Cor.' : '67',
        '2Cor.' : '68',
        'Gal.'  : '69',
        'Eph.'  : '70',
        'Php.'  : '71',
        'Col.'  : '72',
        '1Th.'  : '73',
        '2Th.'  : '74',
        '1Ti.'  : '75',
        '2Ti.'  : '76',
        'Tit.'  : '77',
        'Phm.'  : '78',
        'Heb.'  : '79',
        'Jas.'  : '80',
        '1Pe.'  : '81',
        '2Pe.'  : '82',
        '1Jn.'  : '83',
        '2Jn.'  : '84',
        '3Jn.'  : '85',
        'Jude'  : '86',
        'Rev.'  : '87'

        }


    # Formal long names.
    BOOKNAMESLONG = [

        'Genesis',
        'Exodus',
        'Leviticus',
        'Numbers', 
        'Deuteronomy',
        'Joshua',
        'Judges',
        'Ruth',
        '1 Samuel',
        '2 Samuel',
        '1 Kings',
        '2 Kings', 
        '1 Chronicles',
        '2 Chronicles',
        'Ezra',
        'Nehemiah', 
        'Esther',
        'Job',
        'Psalms',
        'Proverbs', 
        'Ecclesiastes',
        'Song of Solomon',
        'Isaiah',
        'Jeremiah', 
        'Lamentations',
        'Ezekiel',
        'Daniel',
        'Hosea', 
        'Joel',
        'Amos',
        'Obadiah',
        'Jonah', 
        'Micah',
        'Nahum',
        'Habbakkuk',
        'Zephaniah', 
        'Haggai',
        'Zechariah',
        'Malachi',
        'Matthew', 
        'Mark',
        'Luke',
        'John',
        'Acts', 
        'Romans',
        '1 Corinthians',
        '2 Corinthians',
        'Galatians', 
        'Ephesians',
        'Philippians',
        'Colossians',
        '1 Thessalonians', 
        '2 Thessalonians',
        '1 Timothy',
        '2 Timothy',
        'Titus', 
        'Philemon',
        'Hebrews',
        'James',
        '1 Peter', 
        '2 Peter',
        '1 John',
        '2 John',
        '3 John', 
        'Jude',
        'Revelation'

        ]

    # Libronix Generalized Markup (LGM) Booknames.
    BOOKNAMESLGM = [

        'Ge',
        'Ex',
        'Le',
        'Nu',
        'Dt',
        'Jos',
        'Jdg',
        'Ru',
        '1S',
        '2S',
        '1K',
        '2K',
        '1Ch',
        '2Ch',
        'Ezr',
        'Ne',
        'Es',
        'Job',
        'Ps',
        'Pr',
        'Ec',
        'Song',
        'Isa',
        'Jer',
        'Lam',
        'Eze',
        'Dan',
        'Hos',
        'Joel',
        'Amos',
        'Ob',
        'Jonah',
        'Mic',
        'Nah',
        'Hab',
        'Zep',
        'Hag',
        'Zec',
        'Mal',
        'Mt',
        'Mk',
        'Lk',
        'Jn',
        'Acts',
        'Ro',
        '1Co',
        '2Co',
        'Gal',
        'Eph',
        'Php',
        'Col',
        '1Th',
        '2Th',
        '1Ti',
        '2Ti',
        'Tit',
        'Phm',
        'Heb',
        'Jas',
        '1Pe',
        '2Pe',
        '1Jn',
        '2Jn',
        '3Jn',
        'Jude',
        'Re',

        ]

    # Valid book names and chapter count.
    BOOKS = {

        # Bible
        'Gen.' : 50, 'Ex.' : 40, 'Lev.' : 27, 'Num.' : 36, 
        'Deu.' : 34, 'Jos.' : 24, 'Jdg.' : 21, 'Ru.' : 4, 
        '1S.' : 31, '2S.' : 24, '1K.' : 22, '2K.' : 25, 
        '1Chr.' : 29, '2Chr.' : 36, 'Ezra' : 10, 'Ne.' : 13, 
        'Est.' : 10, 'Job' : 42, 'Ps.' : 150, 'Pr.' : 31, 
        'Ecc.' : 12, 'Sos.' : 8, 'Isa.' : 66, 'Jer.' : 52, 
        'Lam.' : 5, 'Eze.' : 48, 'Dan.' : 12, 'Hos.' : 14, 
        'Joel' : 3, 'Amos' : 9, 'Ob.' : 1, 'Jonah' : 4, 
        'Mic.' : 7, 'Nah.' : 3, 'Hab.' : 3, 'Zep.' : 3, 
        'Hag.' : 2, 'Zec.' : 14, 'Mal.' : 4, 'Mat.' : 28, 
        'Mark' : 16, 'Luke' : 24, 'John' : 21, 'Acts' : 28, 
        'Rom.' : 16, '1Cor.' : 16, '2Cor.' : 13, 'Gal.' : 6, 
        'Eph.' : 6, 'Php.' : 4, 'Col.' : 4, '1Th.' : 5, 
        '2Th.' : 3, '1Ti.' : 6, '2Ti.' : 4, 'Tit.' : 3, 
        'Phm.' : 1, 'Heb.' : 13, 'Jas.' : 5, '1Pe.' : 5, 
        '2Pe.' : 3, '1Jn.' : 5, '2Jn.' : 1, '3Jn.' : 1, 
        'Jude' : 1, 'Rev.' : 22,

        # Koran
        #'Surah' : 114
        }

    # Provides number of verses for each chapter and number of chapters
    # for each book.
    RANGES = {
    'Gen.1' : 31,
    'Gen.2' : 25,
    'Gen.3' : 24,
    'Gen.4' : 26,
    'Gen.5' : 32,
    'Gen.6' : 22,
    'Gen.7' : 24,
    'Gen.8' : 22,
    'Gen.9' : 29,
    'Gen.10' : 32,
    'Gen.11' : 32,
    'Gen.12' : 20,
    'Gen.13' : 18,
    'Gen.14' : 24,
    'Gen.15' : 21,
    'Gen.16' : 16,
    'Gen.17' : 27,
    'Gen.18' : 33,
    'Gen.19' : 38,
    'Gen.20' : 18,
    'Gen.21' : 34,
    'Gen.22' : 24,
    'Gen.23' : 20,
    'Gen.24' : 67,
    'Gen.25' : 34,
    'Gen.26' : 35,
    'Gen.27' : 46,
    'Gen.28' : 22,
    'Gen.29' : 35,
    'Gen.30' : 43,
    'Gen.31' : 55,
    'Gen.32' : 32,
    'Gen.33' : 20,
    'Gen.34' : 31,
    'Gen.35' : 29,
    'Gen.36' : 43,
    'Gen.37' : 36,
    'Gen.38' : 30,
    'Gen.39' : 23,
    'Gen.40' : 23,
    'Gen.41' : 57,
    'Gen.42' : 38,
    'Gen.43' : 34,
    'Gen.44' : 34,
    'Gen.45' : 28,
    'Gen.46' : 34,
    'Gen.47' : 31,
    'Gen.48' : 22,
    'Gen.49' : 33,
    'Gen.50' : 26,
    'Gen.' : 50,
    'Ex.1' : 22,
    'Ex.2' : 25,
    'Ex.3' : 22,
    'Ex.4' : 31,
    'Ex.5' : 23,
    'Ex.6' : 30,
    'Ex.7' : 25,
    'Ex.8' : 32,
    'Ex.9' : 35,
    'Ex.10' : 29,
    'Ex.11' : 10,
    'Ex.12' : 51,
    'Ex.13' : 22,
    'Ex.14' : 31,
    'Ex.15' : 27,
    'Ex.16' : 36,
    'Ex.17' : 16,
    'Ex.18' : 27,
    'Ex.19' : 25,
    'Ex.20' : 26,
    'Ex.21' : 36,
    'Ex.22' : 31,
    'Ex.23' : 33,
    'Ex.24' : 18,
    'Ex.25' : 40,
    'Ex.26' : 37,
    'Ex.27' : 21,
    'Ex.28' : 43,
    'Ex.29' : 46,
    'Ex.30' : 38,
    'Ex.31' : 18,
    'Ex.32' : 35,
    'Ex.33' : 23,
    'Ex.34' : 35,
    'Ex.35' : 35,
    'Ex.36' : 38,
    'Ex.37' : 29,
    'Ex.38' : 31,
    'Ex.39' : 43,
    'Ex.40' : 38,
    'Ex.' : 40,
    'Lev.1' : 17,
    'Lev.2' : 16,
    'Lev.3' : 17,
    'Lev.4' : 35,
    'Lev.5' : 19,
    'Lev.6' : 30,
    'Lev.7' : 38,
    'Lev.8' : 36,
    'Lev.9' : 24,
    'Lev.10' : 20,
    'Lev.11' : 47,
    'Lev.12' : 8,
    'Lev.13' : 59,
    'Lev.14' : 57,
    'Lev.15' : 33,
    'Lev.16' : 34,
    'Lev.17' : 16,
    'Lev.18' : 30,
    'Lev.19' : 37,
    'Lev.20' : 27,
    'Lev.21' : 24,
    'Lev.22' : 33,
    'Lev.23' : 44,
    'Lev.24' : 23,
    'Lev.25' : 55,
    'Lev.26' : 46,
    'Lev.27' : 34,
    'Lev.' : 27,
    'Num.1' : 54,
    'Num.2' : 34,
    'Num.3' : 51,
    'Num.4' : 49,
    'Num.5' : 31,
    'Num.6' : 27,
    'Num.7' : 89,
    'Num.8' : 26,
    'Num.9' : 23,
    'Num.10' : 36,
    'Num.11' : 35,
    'Num.12' : 16,
    'Num.13' : 33,
    'Num.14' : 45,
    'Num.15' : 41,
    'Num.16' : 50,
    'Num.17' : 13,
    'Num.18' : 32,
    'Num.19' : 22,
    'Num.20' : 29,
    'Num.21' : 35,
    'Num.22' : 41,
    'Num.23' : 30,
    'Num.24' : 25,
    'Num.25' : 18,
    'Num.26' : 65,
    'Num.27' : 23,
    'Num.28' : 31,
    'Num.29' : 40,
    'Num.30' : 16,
    'Num.31' : 54,
    'Num.32' : 42,
    'Num.33' : 56,
    'Num.34' : 29,
    'Num.35' : 34,
    'Num.36' : 13,
    'Num.' : 36,
    'Deu.1' : 46,
    'Deu.2' : 37,
    'Deu.3' : 29,
    'Deu.4' : 49,
    'Deu.5' : 33,
    'Deu.6' : 25,
    'Deu.7' : 26,
    'Deu.8' : 20,
    'Deu.9' : 29,
    'Deu.10' : 22,
    'Deu.11' : 32,
    'Deu.12' : 32,
    'Deu.13' : 18,
    'Deu.14' : 29,
    'Deu.15' : 23,
    'Deu.16' : 22,
    'Deu.17' : 20,
    'Deu.18' : 22,
    'Deu.19' : 21,
    'Deu.20' : 20,
    'Deu.21' : 23,
    'Deu.22' : 30,
    'Deu.23' : 25,
    'Deu.24' : 22,
    'Deu.25' : 19,
    'Deu.26' : 19,
    'Deu.27' : 26,
    'Deu.28' : 68,
    'Deu.29' : 29,
    'Deu.30' : 20,
    'Deu.31' : 30,
    'Deu.32' : 52,
    'Deu.33' : 29,
    'Deu.34' : 12,
    'Deu.' : 34,
    'Jos.1' : 18,
    'Jos.2' : 24,
    'Jos.3' : 17,
    'Jos.4' : 24,
    'Jos.5' : 15,
    'Jos.6' : 27,
    'Jos.7' : 26,
    'Jos.8' : 35,
    'Jos.9' : 27,
    'Jos.10' : 43,
    'Jos.11' : 23,
    'Jos.12' : 24,
    'Jos.13' : 33,
    'Jos.14' : 15,
    'Jos.15' : 63,
    'Jos.16' : 10,
    'Jos.17' : 18,
    'Jos.18' : 28,
    'Jos.19' : 51,
    'Jos.20' : 9,
    'Jos.21' : 45,
    'Jos.22' : 34,
    'Jos.23' : 16,
    'Jos.24' : 33,
    'Jos.' : 24,
    'Jdg.1' : 36,
    'Jdg.2' : 23,
    'Jdg.3' : 31,
    'Jdg.4' : 24,
    'Jdg.5' : 31,
    'Jdg.6' : 40,
    'Jdg.7' : 25,
    'Jdg.8' : 35,
    'Jdg.9' : 57,
    'Jdg.10' : 18,
    'Jdg.11' : 40,
    'Jdg.12' : 15,
    'Jdg.13' : 25,
    'Jdg.14' : 20,
    'Jdg.15' : 20,
    'Jdg.16' : 31,
    'Jdg.17' : 13,
    'Jdg.18' : 31,
    'Jdg.19' : 30,
    'Jdg.20' : 48,
    'Jdg.21' : 25,
    'Jdg.' : 21,
    'Ru.1' : 22,
    'Ru.2' : 23,
    'Ru.3' : 18,
    'Ru.4' : 22,
    'Ru.' : 4,
    '1S.1' : 28,
    '1S.2' : 36,
    '1S.3' : 21,
    '1S.4' : 22,
    '1S.5' : 12,
    '1S.6' : 21,
    '1S.7' : 17,
    '1S.8' : 22,
    '1S.9' : 27,
    '1S.10' : 27,
    '1S.11' : 15,
    '1S.12' : 25,
    '1S.13' : 23,
    '1S.14' : 52,
    '1S.15' : 35,
    '1S.16' : 23,
    '1S.17' : 58,
    '1S.18' : 30,
    '1S.19' : 24,
    '1S.20' : 42,
    '1S.21' : 15,
    '1S.22' : 23,
    '1S.23' : 29,
    '1S.24' : 22,
    '1S.25' : 44,
    '1S.26' : 25,
    '1S.27' : 12,
    '1S.28' : 25,
    '1S.29' : 11,
    '1S.30' : 31,
    '1S.31' : 13,
    '1S.' : 31,
    '2S.1' : 27,
    '2S.2' : 32,
    '2S.3' : 39,
    '2S.4' : 12,
    '2S.5' : 25,
    '2S.6' : 23,
    '2S.7' : 29,
    '2S.8' : 18,
    '2S.9' : 13,
    '2S.10' : 19,
    '2S.11' : 27,
    '2S.12' : 31,
    '2S.13' : 39,
    '2S.14' : 33,
    '2S.15' : 37,
    '2S.16' : 23,
    '2S.17' : 29,
    '2S.18' : 33,
    '2S.19' : 43,
    '2S.20' : 26,
    '2S.21' : 22,
    '2S.22' : 51,
    '2S.23' : 39,
    '2S.24' : 25,
    '2S.' : 24,
    '1K.1' : 53,
    '1K.2' : 46,
    '1K.3' : 28,
    '1K.4' : 34,
    '1K.5' : 18,
    '1K.6' : 38,
    '1K.7' : 51,
    '1K.8' : 66,
    '1K.9' : 28,
    '1K.10' : 29,
    '1K.11' : 43,
    '1K.12' : 33,
    '1K.13' : 34,
    '1K.14' : 31,
    '1K.15' : 34,
    '1K.16' : 34,
    '1K.17' : 24,
    '1K.18' : 46,
    '1K.19' : 21,
    '1K.20' : 43,
    '1K.21' : 29,
    '1K.22' : 53,
    '1K.' : 22,
    '2K.1' : 18,
    '2K.2' : 25,
    '2K.3' : 27,
    '2K.4' : 44,
    '2K.5' : 27,
    '2K.6' : 33,
    '2K.7' : 20,
    '2K.8' : 29,
    '2K.9' : 37,
    '2K.10' : 36,
    '2K.11' : 21,
    '2K.12' : 21,
    '2K.13' : 25,
    '2K.14' : 29,
    '2K.15' : 38,
    '2K.16' : 20,
    '2K.17' : 41,
    '2K.18' : 37,
    '2K.19' : 37,
    '2K.20' : 21,
    '2K.21' : 26,
    '2K.22' : 20,
    '2K.23' : 37,
    '2K.24' : 20,
    '2K.25' : 30,
    '2K.' : 25,
    '1Chr.1' : 54,
    '1Chr.2' : 55,
    '1Chr.3' : 24,
    '1Chr.4' : 43,
    '1Chr.5' : 26,
    '1Chr.6' : 81,
    '1Chr.7' : 40,
    '1Chr.8' : 40,
    '1Chr.9' : 44,
    '1Chr.10' : 14,
    '1Chr.11' : 47,
    '1Chr.12' : 40,
    '1Chr.13' : 14,
    '1Chr.14' : 17,
    '1Chr.15' : 29,
    '1Chr.16' : 43,
    '1Chr.17' : 27,
    '1Chr.18' : 17,
    '1Chr.19' : 19,
    '1Chr.20' : 8,
    '1Chr.21' : 30,
    '1Chr.22' : 19,
    '1Chr.23' : 32,
    '1Chr.24' : 31,
    '1Chr.25' : 31,
    '1Chr.26' : 32,
    '1Chr.27' : 34,
    '1Chr.28' : 21,
    '1Chr.29' : 30,
    '1Chr.' : 29,
    '2Chr.1' : 17,
    '2Chr.2' : 18,
    '2Chr.3' : 17,
    '2Chr.4' : 22,
    '2Chr.5' : 14,
    '2Chr.6' : 42,
    '2Chr.7' : 22,
    '2Chr.8' : 18,
    '2Chr.9' : 31,
    '2Chr.10' : 19,
    '2Chr.11' : 23,
    '2Chr.12' : 16,
    '2Chr.13' : 22,
    '2Chr.14' : 15,
    '2Chr.15' : 19,
    '2Chr.16' : 14,
    '2Chr.17' : 19,
    '2Chr.18' : 34,
    '2Chr.19' : 11,
    '2Chr.20' : 37,
    '2Chr.21' : 20,
    '2Chr.22' : 12,
    '2Chr.23' : 21,
    '2Chr.24' : 27,
    '2Chr.25' : 28,
    '2Chr.26' : 23,
    '2Chr.27' : 9,
    '2Chr.28' : 27,
    '2Chr.29' : 36,
    '2Chr.30' : 27,
    '2Chr.31' : 21,
    '2Chr.32' : 33,
    '2Chr.33' : 25,
    '2Chr.34' : 33,
    '2Chr.35' : 27,
    '2Chr.36' : 23,
    '2Chr.' : 36,
    'Ezra1' : 11,
    'Ezra2' : 70,
    'Ezra3' : 13,
    'Ezra4' : 24,
    'Ezra5' : 17,
    'Ezra6' : 22,
    'Ezra7' : 28,
    'Ezra8' : 36,
    'Ezra9' : 15,
    'Ezra10' : 44,
    'Ezra' : 10,
    'Ne.1' : 11,
    'Ne.2' : 20,
    'Ne.3' : 32,
    'Ne.4' : 23,
    'Ne.5' : 19,
    'Ne.6' : 19,
    'Ne.7' : 73,
    'Ne.8' : 18,
    'Ne.9' : 38,
    'Ne.10' : 39,
    'Ne.11' : 36,
    'Ne.12' : 47,
    'Ne.13' : 31,
    'Ne.' : 13,
    'Est.1' : 22,
    'Est.2' : 23,
    'Est.3' : 15,
    'Est.4' : 17,
    'Est.5' : 14,
    'Est.6' : 14,
    'Est.7' : 10,
    'Est.8' : 17,
    'Est.9' : 32,
    'Est.10' : 3,
    'Est.' : 10,
    'Job1' : 22,
    'Job2' : 13,
    'Job3' : 26,
    'Job4' : 21,
    'Job5' : 27,
    'Job6' : 30,
    'Job7' : 21,
    'Job8' : 22,
    'Job9' : 35,
    'Job10' : 22,
    'Job11' : 20,
    'Job12' : 25,
    'Job13' : 28,
    'Job14' : 22,
    'Job15' : 35,
    'Job16' : 22,
    'Job17' : 16,
    'Job18' : 21,
    'Job19' : 29,
    'Job20' : 29,
    'Job21' : 34,
    'Job22' : 30,
    'Job23' : 17,
    'Job24' : 25,
    'Job25' : 6,
    'Job26' : 14,
    'Job27' : 23,
    'Job28' : 28,
    'Job29' : 25,
    'Job30' : 31,
    'Job31' : 40,
    'Job32' : 22,
    'Job33' : 33,
    'Job34' : 37,
    'Job35' : 16,
    'Job36' : 33,
    'Job37' : 24,
    'Job38' : 41,
    'Job39' : 30,
    'Job40' : 24,
    'Job41' : 34,
    'Job42' : 17,
    'Job' : 42,
    'Ps.1' : 6,
    'Ps.2' : 12,
    'Ps.3' : 8,
    'Ps.4' : 8,
    'Ps.5' : 12,
    'Ps.6' : 10,
    'Ps.7' : 17,
    'Ps.8' : 9,
    'Ps.9' : 20,
    'Ps.10' : 18,
    'Ps.11' : 7,
    'Ps.12' : 8,
    'Ps.13' : 6,
    'Ps.14' : 7,
    'Ps.15' : 5,
    'Ps.16' : 11,
    'Ps.17' : 15,
    'Ps.18' : 50,
    'Ps.19' : 14,
    'Ps.20' : 9,
    'Ps.21' : 13,
    'Ps.22' : 31,
    'Ps.23' : 6,
    'Ps.24' : 10,
    'Ps.25' : 22,
    'Ps.26' : 12,
    'Ps.27' : 14,
    'Ps.28' : 9,
    'Ps.29' : 11,
    'Ps.30' : 12,
    'Ps.31' : 24,
    'Ps.32' : 11,
    'Ps.33' : 22,
    'Ps.34' : 22,
    'Ps.35' : 28,
    'Ps.36' : 12,
    'Ps.37' : 40,
    'Ps.38' : 22,
    'Ps.39' : 13,
    'Ps.40' : 17,
    'Ps.41' : 13,
    'Ps.42' : 11,
    'Ps.43' : 5,
    'Ps.44' : 26,
    'Ps.45' : 17,
    'Ps.46' : 11,
    'Ps.47' : 9,
    'Ps.48' : 14,
    'Ps.49' : 20,
    'Ps.50' : 23,
    'Ps.51' : 19,
    'Ps.52' : 9,
    'Ps.53' : 6,
    'Ps.54' : 7,
    'Ps.55' : 23,
    'Ps.56' : 13,
    'Ps.57' : 11,
    'Ps.58' : 11,
    'Ps.59' : 17,
    'Ps.60' : 12,
    'Ps.61' : 8,
    'Ps.62' : 12,
    'Ps.63' : 11,
    'Ps.64' : 10,
    'Ps.65' : 13,
    'Ps.66' : 20,
    'Ps.67' : 7,
    'Ps.68' : 35,
    'Ps.69' : 36,
    'Ps.70' : 5,
    'Ps.71' : 24,
    'Ps.72' : 20,
    'Ps.73' : 28,
    'Ps.74' : 23,
    'Ps.75' : 10,
    'Ps.76' : 12,
    'Ps.77' : 20,
    'Ps.78' : 72,
    'Ps.79' : 13,
    'Ps.80' : 19,
    'Ps.81' : 16,
    'Ps.82' : 8,
    'Ps.83' : 18,
    'Ps.84' : 12,
    'Ps.85' : 13,
    'Ps.86' : 17,
    'Ps.87' : 7,
    'Ps.88' : 18,
    'Ps.89' : 52,
    'Ps.90' : 17,
    'Ps.91' : 16,
    'Ps.92' : 15,
    'Ps.93' : 5,
    'Ps.94' : 23,
    'Ps.95' : 11,
    'Ps.96' : 13,
    'Ps.97' : 12,
    'Ps.98' : 9,
    'Ps.99' : 9,
    'Ps.100' : 5,
    'Ps.101' : 8,
    'Ps.102' : 28,
    'Ps.103' : 22,
    'Ps.104' : 35,
    'Ps.105' : 45,
    'Ps.106' : 48,
    'Ps.107' : 43,
    'Ps.108' : 13,
    'Ps.109' : 31,
    'Ps.110' : 7,
    'Ps.111' : 10,
    'Ps.112' : 10,
    'Ps.113' : 9,
    'Ps.114' : 8,
    'Ps.115' : 18,
    'Ps.116' : 19,
    'Ps.117' : 2,
    'Ps.118' : 29,
    'Ps.119' : 176,
    'Ps.120' : 7,
    'Ps.121' : 8,
    'Ps.122' : 9,
    'Ps.123' : 4,
    'Ps.124' : 8,
    'Ps.125' : 5,
    'Ps.126' : 6,
    'Ps.127' : 5,
    'Ps.128' : 6,
    'Ps.129' : 8,
    'Ps.130' : 8,
    'Ps.131' : 3,
    'Ps.132' : 18,
    'Ps.133' : 3,
    'Ps.134' : 3,
    'Ps.135' : 21,
    'Ps.136' : 26,
    'Ps.137' : 9,
    'Ps.138' : 8,
    'Ps.139' : 24,
    'Ps.140' : 13,
    'Ps.141' : 10,
    'Ps.142' : 7,
    'Ps.143' : 12,
    'Ps.144' : 15,
    'Ps.145' : 21,
    'Ps.146' : 10,
    'Ps.147' : 20,
    'Ps.148' : 14,
    'Ps.149' : 9,
    'Ps.150' : 6,
    'Ps.' : 150,
    'Pr.1' : 33,
    'Pr.2' : 22,
    'Pr.3' : 35,
    'Pr.4' : 27,
    'Pr.5' : 23,
    'Pr.6' : 35,
    'Pr.7' : 27,
    'Pr.8' : 36,
    'Pr.9' : 18,
    'Pr.10' : 32,
    'Pr.11' : 31,
    'Pr.12' : 28,
    'Pr.13' : 25,
    'Pr.14' : 35,
    'Pr.15' : 33,
    'Pr.16' : 33,
    'Pr.17' : 28,
    'Pr.18' : 24,
    'Pr.19' : 29,
    'Pr.20' : 30,
    'Pr.21' : 31,
    'Pr.22' : 29,
    'Pr.23' : 35,
    'Pr.24' : 34,
    'Pr.25' : 28,
    'Pr.26' : 28,
    'Pr.27' : 27,
    'Pr.28' : 28,
    'Pr.29' : 27,
    'Pr.30' : 33,
    'Pr.31' : 31,
    'Pr.' : 31,
    'Ecc.1' : 18,
    'Ecc.2' : 26,
    'Ecc.3' : 22,
    'Ecc.4' : 16,
    'Ecc.5' : 20,
    'Ecc.6' : 12,
    'Ecc.7' : 29,
    'Ecc.8' : 17,
    'Ecc.9' : 18,
    'Ecc.10' : 20,
    'Ecc.11' : 10,
    'Ecc.12' : 14,
    'Ecc.' : 12,
    'Sos.1' : 17,
    'Sos.2' : 17,
    'Sos.3' : 11,
    'Sos.4' : 16,
    'Sos.5' : 16,
    'Sos.6' : 13,
    'Sos.7' : 13,
    'Sos.8' : 14,
    'Sos.' : 8,
    'Isa.1' : 31,
    'Isa.2' : 22,
    'Isa.3' : 26,
    'Isa.4' : 6,
    'Isa.5' : 30,
    'Isa.6' : 13,
    'Isa.7' : 25,
    'Isa.8' : 22,
    'Isa.9' : 21,
    'Isa.10' : 34,
    'Isa.11' : 16,
    'Isa.12' : 6,
    'Isa.13' : 22,
    'Isa.14' : 32,
    'Isa.15' : 9,
    'Isa.16' : 14,
    'Isa.17' : 14,
    'Isa.18' : 7,
    'Isa.19' : 25,
    'Isa.20' : 6,
    'Isa.21' : 17,
    'Isa.22' : 25,
    'Isa.23' : 18,
    'Isa.24' : 23,
    'Isa.25' : 12,
    'Isa.26' : 21,
    'Isa.27' : 13,
    'Isa.28' : 29,
    'Isa.29' : 24,
    'Isa.30' : 33,
    'Isa.31' : 9,
    'Isa.32' : 20,
    'Isa.33' : 24,
    'Isa.34' : 17,
    'Isa.35' : 10,
    'Isa.36' : 22,
    'Isa.37' : 38,
    'Isa.38' : 22,
    'Isa.39' : 8,
    'Isa.40' : 31,
    'Isa.41' : 29,
    'Isa.42' : 25,
    'Isa.43' : 28,
    'Isa.44' : 28,
    'Isa.45' : 25,
    'Isa.46' : 13,
    'Isa.47' : 15,
    'Isa.48' : 22,
    'Isa.49' : 26,
    'Isa.50' : 11,
    'Isa.51' : 23,
    'Isa.52' : 15,
    'Isa.53' : 12,
    'Isa.54' : 17,
    'Isa.55' : 13,
    'Isa.56' : 12,
    'Isa.57' : 21,
    'Isa.58' : 14,
    'Isa.59' : 21,
    'Isa.60' : 22,
    'Isa.61' : 11,
    'Isa.62' : 12,
    'Isa.63' : 19,
    'Isa.64' : 12,
    'Isa.65' : 25,
    'Isa.66' : 24,
    'Isa.' : 66,
    'Jer.1' : 19,
    'Jer.2' : 37,
    'Jer.3' : 25,
    'Jer.4' : 31,
    'Jer.5' : 31,
    'Jer.6' : 30,
    'Jer.7' : 34,
    'Jer.8' : 22,
    'Jer.9' : 26,
    'Jer.10' : 25,
    'Jer.11' : 23,
    'Jer.12' : 17,
    'Jer.13' : 27,
    'Jer.14' : 22,
    'Jer.15' : 21,
    'Jer.16' : 21,
    'Jer.17' : 27,
    'Jer.18' : 23,
    'Jer.19' : 15,
    'Jer.20' : 18,
    'Jer.21' : 14,
    'Jer.22' : 30,
    'Jer.23' : 40,
    'Jer.24' : 10,
    'Jer.25' : 38,
    'Jer.26' : 24,
    'Jer.27' : 22,
    'Jer.28' : 17,
    'Jer.29' : 32,
    'Jer.30' : 24,
    'Jer.31' : 40,
    'Jer.32' : 44,
    'Jer.33' : 26,
    'Jer.34' : 22,
    'Jer.35' : 19,
    'Jer.36' : 32,
    'Jer.37' : 21,
    'Jer.38' : 28,
    'Jer.39' : 18,
    'Jer.40' : 16,
    'Jer.41' : 18,
    'Jer.42' : 22,
    'Jer.43' : 13,
    'Jer.44' : 30,
    'Jer.45' : 5,
    'Jer.46' : 28,
    'Jer.47' : 7,
    'Jer.48' : 47,
    'Jer.49' : 39,
    'Jer.50' : 46,
    'Jer.51' : 64,
    'Jer.52' : 34,
    'Jer.' : 52,
    'Lam.1' : 22,
    'Lam.2' : 22,
    'Lam.3' : 66,
    'Lam.4' : 22,
    'Lam.5' : 22,
    'Lam.' : 5,
    'Eze.1' : 28,
    'Eze.2' : 10,
    'Eze.3' : 27,
    'Eze.4' : 17,
    'Eze.5' : 17,
    'Eze.6' : 14,
    'Eze.7' : 27,
    'Eze.8' : 18,
    'Eze.9' : 11,
    'Eze.10' : 22,
    'Eze.11' : 25,
    'Eze.12' : 28,
    'Eze.13' : 23,
    'Eze.14' : 23,
    'Eze.15' : 8,
    'Eze.16' : 63,
    'Eze.17' : 24,
    'Eze.18' : 32,
    'Eze.19' : 14,
    'Eze.20' : 49,
    'Eze.21' : 32,
    'Eze.22' : 31,
    'Eze.23' : 49,
    'Eze.24' : 27,
    'Eze.25' : 17,
    'Eze.26' : 21,
    'Eze.27' : 36,
    'Eze.28' : 26,
    'Eze.29' : 21,
    'Eze.30' : 26,
    'Eze.31' : 18,
    'Eze.32' : 32,
    'Eze.33' : 33,
    'Eze.34' : 31,
    'Eze.35' : 15,
    'Eze.36' : 38,
    'Eze.37' : 28,
    'Eze.38' : 23,
    'Eze.39' : 29,
    'Eze.40' : 49,
    'Eze.41' : 26,
    'Eze.42' : 20,
    'Eze.43' : 27,
    'Eze.44' : 31,
    'Eze.45' : 25,
    'Eze.46' : 24,
    'Eze.47' : 23,
    'Eze.48' : 35,
    'Eze.' : 48,
    'Dan.1' : 21,
    'Dan.2' : 49,
    'Dan.3' : 30,
    'Dan.4' : 37,
    'Dan.5' : 31,
    'Dan.6' : 28,
    'Dan.7' : 28,
    'Dan.8' : 27,
    'Dan.9' : 27,
    'Dan.10' : 21,
    'Dan.11' : 45,
    'Dan.12' : 13,
    'Dan.' : 12,
    'Hos.1' : 11,
    'Hos.2' : 23,
    'Hos.3' : 5,
    'Hos.4' : 19,
    'Hos.5' : 15,
    'Hos.6' : 11,
    'Hos.7' : 16,
    'Hos.8' : 14,
    'Hos.9' : 17,
    'Hos.10' : 15,
    'Hos.11' : 12,
    'Hos.12' : 14,
    'Hos.13' : 16,
    'Hos.14' : 9,
    'Hos.' : 14,
    'Joel1' : 20,
    'Joel2' : 32,
    'Joel3' : 21,
    'Joel' : 3,
    'Amos1' : 15,
    'Amos2' : 16,
    'Amos3' : 15,
    'Amos4' : 13,
    'Amos5' : 27,
    'Amos6' : 14,
    'Amos7' : 17,
    'Amos8' : 14,
    'Amos9' : 15,
    'Amos' : 9,
    'Ob.1' : 21,
    'Ob.' : 1,
    'Jonah1' : 17,
    'Jonah2' : 10,
    'Jonah3' : 10,
    'Jonah4' : 11,
    'Jonah' : 4,
    'Mic.1' : 16,
    'Mic.2' : 13,
    'Mic.3' : 12,
    'Mic.4' : 13,
    'Mic.5' : 15,
    'Mic.6' : 16,
    'Mic.7' : 20,
    'Mic.' : 7,
    'Nah.1' : 15,
    'Nah.2' : 13,
    'Nah.3' : 19,
    'Nah.' : 3,
    'Hab.1' : 17,
    'Hab.2' : 20,
    'Hab.3' : 19,
    'Hab.' : 3,
    'Zep.1' : 18,
    'Zep.2' : 15,
    'Zep.3' : 20,
    'Zep.' : 3,
    'Hag.1' : 15,
    'Hag.2' : 23,
    'Hag.' : 2,
    'Zec.1' : 21,
    'Zec.2' : 13,
    'Zec.3' : 10,
    'Zec.4' : 14,
    'Zec.5' : 11,
    'Zec.6' : 15,
    'Zec.7' : 14,
    'Zec.8' : 23,
    'Zec.9' : 17,
    'Zec.10' : 12,
    'Zec.11' : 17,
    'Zec.12' : 14,
    'Zec.13' : 9,
    'Zec.14' : 21,
    'Zec.' : 14,
    'Mal.1' : 14,
    'Mal.2' : 17,
    'Mal.3' : 18,
    'Mal.4' : 6,
    'Mal.' : 4,
    'Mat.1' : 25,
    'Mat.2' : 23,
    'Mat.3' : 17,
    'Mat.4' : 25,
    'Mat.5' : 48,
    'Mat.6' : 34,
    'Mat.7' : 29,
    'Mat.8' : 34,
    'Mat.9' : 38,
    'Mat.10' : 42,
    'Mat.11' : 30,
    'Mat.12' : 50,
    'Mat.13' : 58,
    'Mat.14' : 36,
    'Mat.15' : 39,
    'Mat.16' : 28,
    'Mat.17' : 27,
    'Mat.18' : 35,
    'Mat.19' : 30,
    'Mat.20' : 34,
    'Mat.21' : 46,
    'Mat.22' : 46,
    'Mat.23' : 39,
    'Mat.24' : 51,
    'Mat.25' : 46,
    'Mat.26' : 75,
    'Mat.27' : 66,
    'Mat.28' : 20,
    'Mat.' : 28,
    'Mark1' : 45,
    'Mark2' : 28,
    'Mark3' : 35,
    'Mark4' : 41,
    'Mark5' : 43,
    'Mark6' : 56,
    'Mark7' : 37,
    'Mark8' : 38,
    'Mark9' : 50,
    'Mark10' : 52,
    'Mark11' : 33,
    'Mark12' : 44,
    'Mark13' : 37,
    'Mark14' : 72,
    'Mark15' : 47,
    'Mark16' : 20,
    'Mark' : 16,
    'Luke1' : 80,
    'Luke2' : 52,
    'Luke3' : 38,
    'Luke4' : 44,
    'Luke5' : 39,
    'Luke6' : 49,
    'Luke7' : 50,
    'Luke8' : 56,
    'Luke9' : 62,
    'Luke10' : 42,
    'Luke11' : 54,
    'Luke12' : 59,
    'Luke13' : 35,
    'Luke14' : 35,
    'Luke15' : 32,
    'Luke16' : 31,
    'Luke17' : 37,
    'Luke18' : 43,
    'Luke19' : 48,
    'Luke20' : 47,
    'Luke21' : 38,
    'Luke22' : 71,
    'Luke23' : 56,
    'Luke24' : 53,
    'Luke' : 24,
    'John1' : 51,
    'John2' : 25,
    'John3' : 36,
    'John4' : 54,
    'John5' : 47,
    'John6' : 71,
    'John7' : 53,
    'John8' : 59,
    'John9' : 41,
    'John10' : 42,
    'John11' : 57,
    'John12' : 50,
    'John13' : 38,
    'John14' : 31,
    'John15' : 27,
    'John16' : 33,
    'John17' : 26,
    'John18' : 40,
    'John19' : 42,
    'John20' : 31,
    'John21' : 25,
    'John' : 21,
    'Acts1' : 26,
    'Acts2' : 47,
    'Acts3' : 26,
    'Acts4' : 37,
    'Acts5' : 42,
    'Acts6' : 15,
    'Acts7' : 60,
    'Acts8' : 40,
    'Acts9' : 43,
    'Acts10' : 48,
    'Acts11' : 30,
    'Acts12' : 25,
    'Acts13' : 52,
    'Acts14' : 28,
    'Acts15' : 41,
    'Acts16' : 40,
    'Acts17' : 34,
    'Acts18' : 28,
    'Acts19' : 41,
    'Acts20' : 38,
    'Acts21' : 40,
    'Acts22' : 30,
    'Acts23' : 35,
    'Acts24' : 27,
    'Acts25' : 27,
    'Acts26' : 32,
    'Acts27' : 44,
    'Acts28' : 31,
    'Acts' : 28,
    'Rom.1' : 32,
    'Rom.2' : 29,
    'Rom.3' : 31,
    'Rom.4' : 25,
    'Rom.5' : 21,
    'Rom.6' : 23,
    'Rom.7' : 25,
    'Rom.8' : 39,
    'Rom.9' : 33,
    'Rom.10' : 21,
    'Rom.11' : 36,
    'Rom.12' : 21,
    'Rom.13' : 14,
    'Rom.14' : 23,
    'Rom.15' : 33,
    'Rom.16' : 27,
    'Rom.' : 16,
    '1Cor.1' : 31,
    '1Cor.2' : 16,
    '1Cor.3' : 23,
    '1Cor.4' : 21,
    '1Cor.5' : 13,
    '1Cor.6' : 20,
    '1Cor.7' : 40,
    '1Cor.8' : 13,
    '1Cor.9' : 27,
    '1Cor.10' : 33,
    '1Cor.11' : 34,
    '1Cor.12' : 31,
    '1Cor.13' : 13,
    '1Cor.14' : 40,
    '1Cor.15' : 58,
    '1Cor.16' : 24,
    '1Cor.' : 16,
    '2Cor.1' : 24,
    '2Cor.2' : 17,
    '2Cor.3' : 18,
    '2Cor.4' : 18,
    '2Cor.5' : 21,
    '2Cor.6' : 18,
    '2Cor.7' : 16,
    '2Cor.8' : 24,
    '2Cor.9' : 15,
    '2Cor.10' : 18,
    '2Cor.11' : 33,
    '2Cor.12' : 21,
    '2Cor.13' : 14,
    '2Cor.' : 13,
    'Gal.1' : 24,
    'Gal.2' : 21,
    'Gal.3' : 29,
    'Gal.4' : 31,
    'Gal.5' : 26,
    'Gal.6' : 18,
    'Gal.' : 6,
    'Eph.1' : 23,
    'Eph.2' : 22,
    'Eph.3' : 21,
    'Eph.4' : 32,
    'Eph.5' : 33,
    'Eph.6' : 24,
    'Eph.' : 6,
    'Php.1' : 30,
    'Php.2' : 30,
    'Php.3' : 21,
    'Php.4' : 23,
    'Php.' : 4,
    'Col.1' : 29,
    'Col.2' : 23,
    'Col.3' : 25,
    'Col.4' : 18,
    'Col.' : 4,
    '1Th.1' : 10,
    '1Th.2' : 20,
    '1Th.3' : 13,
    '1Th.4' : 18,
    '1Th.5' : 28,
    '1Th.' : 5,
    '2Th.1' : 12,
    '2Th.2' : 17,
    '2Th.3' : 18,
    '2Th.' : 3,
    '1Ti.1' : 20,
    '1Ti.2' : 15,
    '1Ti.3' : 16,
    '1Ti.4' : 16,
    '1Ti.5' : 25,
    '1Ti.6' : 21,
    '1Ti.' : 6,
    '2Ti.1' : 18,
    '2Ti.2' : 26,
    '2Ti.3' : 17,
    '2Ti.4' : 22,
    '2Ti.' : 4,
    'Tit.1' : 16,
    'Tit.2' : 15,
    'Tit.3' : 15,
    'Tit.' : 3,
    'Phm.1' : 25,
    'Phm.' : 1,
    'Heb.1' : 14,
    'Heb.2' : 18,
    'Heb.3' : 19,
    'Heb.4' : 16,
    'Heb.5' : 14,
    'Heb.6' : 20,
    'Heb.7' : 28,
    'Heb.8' : 13,
    'Heb.9' : 28,
    'Heb.10' : 39,
    'Heb.11' : 40,
    'Heb.12' : 29,
    'Heb.13' : 25,
    'Heb.' : 13,
    'Jas.1' : 27,
    'Jas.2' : 26,
    'Jas.3' : 18,
    'Jas.4' : 17,
    'Jas.5' : 20,
    'Jas.' : 5,
    '1Pe.1' : 25,
    '1Pe.2' : 25,
    '1Pe.3' : 22,
    '1Pe.4' : 19,
    '1Pe.5' : 14,
    '1Pe.' : 5,
    '2Pe.1' : 21,
    '2Pe.2' : 22,
    '2Pe.3' : 18,
    '2Pe.' : 3,
    '1Jn.1' : 10,
    '1Jn.2' : 29,
    '1Jn.3' : 24,
    '1Jn.4' : 21,
    '1Jn.5' : 21,
    '1Jn.' : 5,
    '2Jn.1' : 13,
    '2Jn.' : 1,
    '3Jn.1' : 14,
    '3Jn.' : 1,
    'Jude1' : 25,
    'Jude' : 1,
    'Rev.1' : 20,
    'Rev.2' : 29,
    'Rev.3' : 22,
    'Rev.4' : 11,
    'Rev.5' : 14,
    'Rev.6' : 17,
    'Rev.7' : 17,
    'Rev.8' : 13,
    'Rev.9' : 21,
    'Rev.10' : 11,
    'Rev.11' : 19,
    'Rev.12' : 17,
    'Rev.13' : 18,
    'Rev.14' : 20,
    'Rev.15' : 8,
    'Rev.16' : 21,
    'Rev.17' : 18,
    'Rev.18' : 24,
    'Rev.19' : 21,
    'Rev.20' : 15,
    'Rev.21' : 27,
    'Rev.22' : 21,
    'Rev.' :  22
    }

    def __init__( self, book, chapter, verse, start = -1, end = -1, text = '' ) :
        # Initialize an address object.

        # Book name and index.
        self.m_book = book
        self.m_iBook = self.iBook( self.m_book )
        self.m_bookLong = self.BOOKNAMESLONG[ self.m_iBook ]

        # Chapter.
        try :
            self.m_chapter = chapter
            iChapter = eval( chapter )
            #maxChapter = self.BOOKS[ self.m_book ]
            maxChapter = self.BOOKS[ self.BOOKNAMES[ self.m_iBook ] ]
            assert( 0 < iChapter and iChapter <= maxChapter )
        except :
            msg = 'Invalid chapter value in %s: %s' % (self.m_book, self.m_chapter )
            raise AddressError, msg

        self.m_verseRange = verse
        parts = self.m_verseRange.split( '-' )
        self.m_verse = parts[ 0 ]

        # Optional text which corresponds to the address.
        self.m_start = start
        self.m_end = end
        self.m_text = text

        # Fabricate our representation if we weren't given one.
        if not self.m_text :
            self.m_text = self.m_book + ' ' + self.m_chapter + ':' + self.m_verseRange

    def dump( self ) :
        print '####### ADDRESS ##########'
        print 'm_book:', self.m_book
        print 'm_chapter:', self.m_chapter
        print 'm_verseRange:', self.m_verseRange
        print 'm_start:', self.m_start
        print 'm_text:', self.m_text
        print 'm_iBook:', self.m_iBook

    def iBook( self, bookname ) :
        # Identify book name offset.
        iBook = None
        try :
            iBook = self.BOOKNAMES.index( bookname )
        except ValueError : 
            try :
                iBook = self.BOOKNAMESLONG.index( bookname )
                bookname = self.BOOKNAMESLONG[ iBook ]
            except ValueError : 
                msg = 'Invalid book name: ' + bookname
                raise AddressError, msg
        return iBook

    def __repr__( self ) :
        # Return the short representation of an address.
        return self.short()

    def length( self ) :
        length = 0
        if -1 == self.m_start or -1 == self.m_end :
            length = len( self.m_text )
        else :
            length = self.m_end - self.m_start
        return length

    def original( self ) :
        text = self.m_text
        return text

    def short( self ) :
        # Return the formal representation, ignoring the original representation.
        text = self.m_book + ' ' + self.m_chapter + ':' + self.m_verseRange
        return text

    def long( self ) :
        # Return the formal long representation, ignoring the original representation.
        text = self.m_bookLong + ' ' + self.m_chapter + ':' + self.m_verseRange
        return text

    def lbxBookNumber( self, bookname = None ) :
        # If not a short bookname.
        if not bookname : bookname = self.book()
        iBook = self.iBook( bookname )
        shortname = self.BOOKNAMES[ iBook ]
        return  self.BOOKNAME_TO_LBX_NUMBER[ shortname ]

    def studylightLink( self, text = None ) :
        if None == text : text = self.m_text
        where = urllib.quote( '%s %s:%s' % ( self.m_bookLong, self.m_chapter, self.m_verseRange ) )

        return '<a title="View passage at www.StudyLight.org" href="https://www.studylight.org/desk/?l=en&query=%s">%s</a>' % ( where, text )

    def gospelcomLink( self, text = None ) :
        if None == text : text = self.m_text
        where = urllib.quote( '%s %s:%s' % ( self.m_bookLong, self.m_chapter, self.m_verseRange ) )

        return '<a title="View passage at bible.GospelCom.net" href="https://bible.gospelcom.net/bible?passage=%s&language=english">%s</a>' % ( where, text )


    def esLink( self, markup = 0 ) :
        # e-Sword format, with optional format markup.
        book = self.BOOKNAMES_ESWORD[ self.m_iBook ]
        text = '%s_%s:%s' % ( book, self.m_chapter, self.m_verseRange )
        if markup : text = r'{\ul\cf11 ' + text + '}'
        return text

    def ssLink( self ) :
        # SwordSearcher format.
        text = '<scripref>%s</scripref>' % self.__repr__()
        return text

    def stLink( self, text = None ) :
        if None == text : text = self.m_text
        where = urllib.quote( '%s %s:%s' % ( self.m_bookLong, self.m_chapter, self.m_verseRange ) )
        return '<a title="View passage at www.SpiritAndTruth.org" href="https://bible.gospelcom.net/bible?passage=%s&language=english">%s</a>' % ( where, text )


    def lbxLaunchLinkImg( self, text = None ) :
        show = '<img border="0" src="LibronixLink.gif"></a>'
        if text : show = text
        return self.lbxLaunchLink( show )
        
    def lbxLaunchLink( self, text = None ) :
        if None == text : text = self.m_text
        where = urllib.quote( '%s %s:%s' % ( self.m_bookLong, self.m_chapter, self.m_verseRange ) )
        return '<a title="View passage in Logos Bible Software Series X" href="libronixdls:keylink|ref=[en]bible:%s">%s</a>' % ( where, text )

    def lbxLink( self, text = None ) :
        # Return our address as a Libronix PBB link.
        if None == text : text = self.m_text
        parts = self.m_verseRange.split( '-' )
        firstVerse = parts[ 0 ]
        lastVerse = ''
        if 1 < len( parts ) : lastVerse = parts[ 1 ]
        firstAddr = '%s.%s.%s' % ( self.lbxBookNumber( self.m_book ), self.m_chapter, firstVerse )
        lastAddr = ''
        addr = firstAddr
        if lastVerse :
            lastAddr = '%s.%s.%s' % ( self.lbxBookNumber( self.m_book ), self.m_chapter, lastVerse )
            addr += '-' + lastAddr
        if text : text += '|'
        link = '[[%sbible.%s]]' % ( text, addr )
        return link

    def lbxMilestone( self ) :
        # Return our address as a Libronix PBB milestone.
        return '[[@bible.%s.%s.%s]]' % ( self.lbxBookNumber( self.m_book ), self.m_chapter, self.m_verse )


    def lgmLink( self, text = None ) :
        # Return our address as a Libronix LGM link.
        lgmBook = self.BOOKNAMESLGM[ self.m_iBook ]
        if None == text : text = self.m_text
        link = '<xbr t="%s %s:%s">%s</xbr>' % ( 
            lgmBook, self.m_chapter, self.m_verseRange, text )
        return link

    def lgmMilestone( self ) :
        # Return our address as a Libronix LGM milestone.
        lgmBook = self.BOOKNAMESLGM[ self.m_iBook ]
        milestone = '<br n="%s %s:%s">' % ( 
            lgmBook, self.m_chapter, self.m_verseRange )
        return milestone


    def book( self ) :
        # Return book name.
        return self.m_book

    def bookNumber( self ) :
        # Return book number, Genesis is book 1.
        return self.m_iBook + 1

    def chapter( self ) :
        return self.m_chapter

    def verseRange( self ) :
        return self.m_verseRange

    def verse( self ) :
        return self.m_verse

    def start( self ) :
        return self.m_start

##########################################################################
# Is Text A Bible Address?
##########################################################################
#
# Makes the simplifying assumption that any link of the form
#
#    <book> <numbers>:<numbers>
#
# is a bible address.
#
def isAddress( text ) :
    
    # Assume not an address.
    found = 0

    # If text contains at least two substrings.
    pieces = string.split( text, ' ' )
    if ( 2 <= len( pieces ) ) :

        # If first substring is a bible book name.
        bookNames = []
        bookNames.extend( Address.BOOKS.keys() )
        bookNames.extend( Address.BOOKNAMESLONG )
        for book in bookNames :
            if ( string.find( pieces[ 0 ], book ) == 0 and
                len( book ) == len( pieces[ 0 ] ) ) :

                # If second substring is an address pattern.
                # Numbers must start second piece.
                match = re.match( Address.ADDRESS_PATTERN, pieces[ 1 ] )
                if match :

                    # If numbers start next piece.
                    if 0 == match.start() :
                        # We have a bible address.
                        found = 1
                    #else :
                    #    print 'IS_ADDRESS_MATCH_FAILURE:', pieces[ 1 ]
                
    #print found, text
    return found

##############################################################################
# Bible Address Finder.
##############################################################################
#
# The pattern matching done within this class is very messy.
# It was originally utilized to convert the International Standard Bible
# Encyclopedia to HTML and has been modified extensively since then.
#
# wHAT IS REALLY NEEDED IS A COMPLETE OVERHAUL.  But there are so many
# minor, complex, and cumulative tweaks in what is currently here that I am loath
# to take the necessary time to revamp the entire approach, as much
# as it cries out for revision.  I'm sure the result would be
# vastly simpler, more reliable, and faster.
#
# Something for another day I guess...
#
class AddressFinder :

    MATCH_DEFAULT = 0
    MATCH_EXACT = 1
    MATCH_IGNORECASE = 2

    def __init__( self, verbose = 0, exactMatch = MATCH_DEFAULT, bookNames = None, assumeBook = '' ) :
        # Initialize AddressFinder
        # Inputs:
        #    Include book name as part of found address text?
        #    Are we to be verbose while matching?
        #       0 = disabled
        #       1 = enabled (using print)
        #       file = log to file
        #    Must bible book names match exactly? Is case important?
        #    Caller may provide optional list of books.
        #    Caller may specify an assumed (default) book.

        matchFlags = 0
        if AddressFinder.MATCH_IGNORECASE == exactMatch : matchFlags |= re.IGNORECASE

        # Allow html to separate book and verses.
        html = '((<[^>]+>)\s*)*'

        # Special conversions for formal book prefixes.
        # Single 'I' pattern REQUIRE a following space so don't get confused
        # with book names such as 'Isa.'
        patternFirst = '^(I\s+|1st|first)\s*'
        patternSecond = '^(II|2nd|second)\s*'
        patternThird = '^(III|3rd|third)\s*'
        self.m_patternFirstExp = re.compile( patternFirst, re.IGNORECASE )
        self.m_patternSecondExp = re.compile( patternSecond, re.IGNORECASE )
        self.m_patternThirdExp = re.compile( patternThird, re.IGNORECASE )

        # Pattern for matching a full address.
        patternBookNumber = r'(?P<number>([1-3]|I{1,3}|[Ff]irst|[Ss]econd|[Tt]hird|1st|2nd|3rd))'
        #patternBookName = r'(?P<book>(' + patternBookNumber + '\s*)?[A-Z][a-z]*\.?)\s+'
        #patternBookName = r'(?P<book>(' + patternBookNumber + '\s*)?[A-Z][a-z]*\.?)'
        # ONLY ALLOW UP TO TWO WHITESPACES BETWEEN BOOK NUMBER AND NAME
        # (OTHERWISE "1              John 1:1" will match as is "1John 1:1")
        patternBookName = r'(?P<book>(' + patternBookNumber + '(\s{1,2})?)?[A-Z][a-z]*\.?)'
        #patternVersesIntro = r'(?P<chapterandverses>[1-9]\d*'
        patternVersesIntro = r'(?P<chapterandverses>[1-9]\d*[abcd]?'

        # Avoid including digits which are the preceeding portion of the next book, but
        # allow single-letter subscripts. Once a single digit is encountered, we accept
        # all consecutive digits except when a colon terminates the sequence.  When
        # spaces occur among digits, then we try to reject inclusion of the book number
        # of any book which may follow. This is admittedly imperfect because we are not
        # looking at the actual book names. For now, we are allowing lowercase bible book
        # names which makes it a bit more difficult.
        #okDigits = r'(\d+(?!(:|\s*[A-Za-z]+\.?)\s+\d|[A-Za-z][A-Za-z]))'
        #okDigits = r'(\d+(?!(:|\s*[A-Za-z]+\.?\s+\d|[A-Za-z][A-Za-z])))'
        #okDigits = r'((?<!\s)\d+(?!\d+:)|\d+(?!(\d+:|\s*[A-Za-z]+\.?\s+\d|[A-Za-z][A-Za-z])))'
        #okDigits = r'((?<!\s)[0-9abcd]+(?!\d+:)|[0-9abcd]+(?!(\d+:|\s*[A-Za-z]+\.?\s+\d|[A-Za-z][A-Za-z])))'
        # ABCD only allowed if immediately preceded by a digit.
        okDigits = r'((?<!\s)[0-9][0-9abcd]*(?!\d+:)|[0-9][0-9abcd]*(?!(\d+:|\s*[A-Za-z]+\.?\s+\d|[A-Za-z][A-Za-z])))'


        #patternVersesBody = r':\d+(\s*[-,;]\s*' + okDigits + ')*'
        #patternVersesBody = r':[0-9a-zA-Z]+(\s*[-,;]\s*' + okDigits + ')*'
        #patternVersesBody = r':[0-9abcd]+(\s*[-,;]\s*' + okDigits + ')*'
        #patternVersesBody = r':[0-9abcd]+(\s*([-,;]|\s+(&|[Aa][Nn][Dd])\s+)\s*' + okDigits + ')*'
        #patternVersesBody = r'(:[0-9abcd]|,)+(\s*([-,;]|\s+(&|[Aa][Nn][Dd])\s+)?\s*' + okDigits + ')*'
        #patternVersesBody = r'(:[0-9abcd]|,|\s+(&|[Aa][Nn][Dd])\s+)+(\s*([-,;]|\s+(&|[Aa][Nn][Dd])\s+)?\s*' + okDigits + ')*'
        #ABCD only allowed if immediately preceded by a digit.
        #patternVersesBody = r'(:[0-9][0-9abcd]?|,|\s+(&|[Aa][Nn][Dd])\s+)+(\s*([-,;]|\s+(&|[Aa][Nn][Dd])\s+)?\s*' + okDigits + ')*'
        #patternVersesBody = r'(:[0-9][0-9abcd]?|,|;|\s+(&|[Aa][Nn][Dd])\s+)+(\s*([-,;]|\s+(&|[Aa][Nn][Dd])\s+)?\s*' + okDigits + ')*'
        patternVersesBody = r'(:[0-9][0-9abcd]?|,|-|;|\s+(&|[Aa][Nn][Dd])\s+)+(\s*([-,;]|\s+(&|[Aa][Nn][Dd])\s+)?\s*' + okDigits + ')*'
        # Allow optional parenthesis or brackets to surround verse range.
        patternVerseRange = '\s+([\(\[]\s*)?' + patternVersesIntro + '(' + patternVersesBody + '(\s*[\)\]])?)*)'
        patternBook = '(' + patternBookName + patternVerseRange + ')'
        self.m_patternBookExp = re.compile( patternBook, matchFlags )

        # Pattern for matching a disconnnected verse specification separated
        # from the book name which appeared earlier in the text.
        #floatingChapterIntro = r'(?P<chapterintro>(\s+([Cc][Hh][Ss]?\.?|[Cc][Hh][Aa][Pp][Tt][Ee][Rr][Ss]?)))?'
        floatingChapterIntro = r'(?P<chapterintro>(\b([Cc][Hh][Ss]?\.?|[Cc][Hh][Aa][Pp][Ss]?\.?|[Cc][Hh][Aa][Pp][Tt][Ee][Rr][Ss]?)))?'
        #floatingVerseIntro = r'(?P<before>(\s+([Vv]{1,2}|[Vv]er)\.|[Vv]erses?|-)\s+)?'
        #floatingVerseIntro = r'(?P<before>(\s+([Vv]{1,2}|[Vv]s|[Vv]er)\.|[Vv]erses?|-|and|&)\s+)?'
        #floatingVerseIntro = r'(?P<before>(\s+([Vv]{1,2}|[Vv]s|[Vv]er)\.|[Vv]erses?|-)\s+)?'
        floatingVerseIntro = r'(?P<verseintro>(\b([Vv]{1,2}|[Vv]s|[Vv][Ee][Rr])\.|[Vv][Ee][Rr][Ss][Ee][Ss]?|-))?'
        floatingVerseOpen = r'\s*(?<=[\[\(\s])'
        #floatingVerseRange = r'(' + floatingChapterIntro + floatingVerseIntro + floatingVerseOpen + patternVersesIntro + r'([^:\s+A-Za-z]|' + patternVersesBody + r')*))'
        floatingVerseRange = r'(' + floatingChapterIntro + floatingVerseIntro + floatingVerseOpen + patternVersesIntro + r'([^:\s+A-Za-z]|\s+(&|[Aa][Nn][Dd])\s+|' + patternVersesBody + r')*))'
        self.m_floatingVerseExp = re.compile( floatingVerseRange, matchFlags )
        
        # Patter for chapter and its associated verses.
        # (Break apart ranges spanning more than a single chapter.)
        #patternChapterVerse = r'\d+:?([0-9]+(?!:)|[,\-\s])*'
        #patternChapterVerse = r'\d+:?([0-9]+(?!\d*:)|[,\-\s])*'
        #patternChapterVerse = r'\d+:?([0-9abcd]+(?!\d*:)|[,\-\s])*'
        #patternChapterVerse = r'\d+:?([0-9a-zA-Z]+(?!\d*:)|[,\-\s])*'
        # Do not include "and" in match.
        #patternChapterVerse = r'\d+:?([0-9abcd]+(?!\d*:)|[,\-\s])*'
        #ABCD only allowed immediately following a digit.
        patternChapterVerse = r'\d+:?([0-9][0-9abcd]*(?!\d*:)|[,\-\s])*'
        self.m_patternChapterVerseExp = re.compile( patternChapterVerse, matchFlags )

        patternChapter = '(?P<chapter>[1-9][0-9]*):?'
        self.m_patternChapterExp = re.compile( patternChapter, matchFlags )

        # Pattern for verses on their own.
        #patternVerse = r'(?P<verse>(?P<first>[1-9]\d*)(\s*-\s*(?P<last>[1-9]\d*))?)[\s,]*'
        #patternVerse = r'(?P<verse>(?P<first>[1-9][0-9a-zA-Z]*)(\s*-\s*(?P<last>[1-9][0-9abcd]*))?)[\s,]*'
        patternVerse = r'(?P<verse>(?P<first>[1-9][0-9abcd]*)(\s*-\s*(?P<last>[1-9][0-9abcd]*))?)[\s,]*'
        self.m_patternVerseExp = re.compile( patternVerse, matchFlags)

        # Punctuation which, if encountered, invalidate the connection between
        # a previously encountered book name and a floating verse range.
        #patternPunc = r'([\(\)\.?])'
        #patternPunc = r'([\(\)\.\?])'
        patternPunc = r'([\)\]\.\?]|\s+\D+\s+\D+\s+)'
        self.m_patternPuncExp = re.compile( patternPunc, matchFlags )


        self.m_verbose = verbose
        self.m_verboseFile = None
        if self.m_verbose and type( self.m_verbose ) != type( 1 ) :
            self.m_verboseFile = self.m_verbose
        self.m_exactMatch = exactMatch
        self.m_bookNames = bookNames
        self.m_assumeBook = assumeBook
        if ( not bookNames ) : self.m_bookNames = []
        self.m_bookNames.extend( Address.BOOKNAMES )
        self.m_bookNames.extend( Address.BOOKNAMESLONG )
        self.m_assumeBookCanonical = self.canonicalBookName( self.m_assumeBook )
        if self.m_assumeBook and not self.m_assumeBookCanonical :
            print 'Unrecognized book name:', self.m_assumeBook
            assert 0

    def findAddresses( self, lineNumber, original, verseRange = 0, includeBookName = 0) :
        # Find bible addresses in a line of text. 
        #
        # Returns a list of address objects representing textual items in the lines.
        #    Each address object includes its original location where found in
        #    the line and the original text it corresponds to.  See the
        #    Address class for additional information.
        # Inputs:
        #    Line number of input (for error messages).
        #    Original line of text.
        #    Chapter without verses use entire verse range?
        #    Override constructor specification to include book name?


        # This was added late in the game after we already set this in the
        # constructor. Rather than changing the constructor signature (or adding
        # another) we allow this to be redundantly specified dynamically by callers
        # of this method--so it can be changed on-the-fly.
        DEBUG = 0
        if DEBUG: 
            print '######## findAddress entry ################'
            print original

        # Start with an empty list.
        addresses = []

        # We operate on shifted copies of the original.
        remainder = original
        lower = original.lower()
        offsetBook = 0
        offsetChapter = 0
        offsetVerse = 0

        bkNamePrevious = None
        matchVerseSpecified = 0
        matchVerseSpecifiedNow = 0
        chapter = ''

        # Look for a book with address range.
        while ( 1 ) :

            #print 'REMAINDER ', remainder
            bkName = None
            matchFull = None
            matchFloating = None
            startFull = startFloating = len( remainder )
            lengthFloating = lengthFull = 0

            # If have seen a previous book or an assumed book name is provided.
            floatingVerse = 0
            if not bkNamePrevious and self.m_assumeBookCanonical : 
                bkNamePrevious = self.m_assumeBookCanonical
            #print 'bkNamePrevious at top', bkNamePrevious
            # Scan for floating addresses until can't find any more candidates.
            iStart = 0
            while( bkNamePrevious ) :

                startPunc = len( remainder )
                matchPunc = self.m_patternPuncExp.search( remainder, iStart )
                if matchPunc :
                    startPunc = matchPunc.start()
                    if DEBUG: print 'startPunc', startPunc, matchPunc.group()

                # Look for floating verse specification.
                confident = 0
                if DEBUG: print 'SEARCHING FOR FLOATING', iStart, remainder[ iStart : iStart + 50 ]
                matchFloating = self.m_floatingVerseExp.search( remainder, iStart )
                if not matchFloating : break
                iStart = matchFloating.end()
                if DEBUG: print 'Line', lineNumber, 'MATCHFLOATING ', matchFloating.start(), matchFloating.group()

                # Did the word prior to the floating address explicitly designate
                # this as a verse?
                matchVerseSpecifiedNow = matchFloating.group( 'verseintro' )
                if matchVerseSpecifiedNow : confident = 1

                # If found chapter/verse separator or explicit introducer, then we are pretty confident that
                # these floating numbers related to the prior book.
                matchChapterSpecified = matchFloating.group( 'chapterintro' )
                chapterAndVerses = matchFloating.group( 'chapterandverses' )
                if matchChapterSpecified or re.match( '\d+:', chapterAndVerses ) :
                    matchVerseSpecified = 0
                    confident = 1
                    # Saw a chapter:verse separator so override explicit designation as verse only.
                    matchVerseSpecifiedNow = 0

                # Latch when we see an explicit verse specifier until we see
                # later format which makes it clear chapters are again in view.
                if DEBUG: print 'MATCH VERSE SPECIFIED #1:', matchVerseSpecified
                matchVerseSpecified = matchVerseSpecifiedNow or matchVerseSpecified
                if DEBUG: print 'MATCH VERSE SPECIFIED #2:', matchVerseSpecified

                # If assumed bookname, then be confident when we see a floating range of numbers.
                if matchFloating  and self.m_assumeBookCanonical :
                    if re.search( '\d\s*-\s*\d', matchFloating.group() ) : confident = 1

                # Assume we will be using the previous book.
                startFloating = matchFloating.start( 'chapterandverses' )
                if DEBUG: print 'FLOATING CONFIDENT & MATCH', confident, matchVerseSpecified, startFloating, bkName

                # Done if no intervening punctuation and floating address is confidently identified.
                if startFloating <= startPunc and confident : break

                # If we have a confident match.
                if confident :

                    # Done if we have a previous bookname or an assumed bookname.
                    if not bkNamePrevious : bkNamePreviuos = self.m_assumeBookCanonical
                    if bkNamePrevious : break

                # Done if we have an assumed book name and a confident match.
                #if self.m_assumeBookCanonical and confident : 
                #    bkNamePrevious = self.m_assumeBookCanonical
                #    break;

                if DEBUG: print 'FLOATING NOT CONFIDENT:', matchFloating.group()

            # While we are finding full match candidates.
            iStart = 0
            while 1 :
                if DEBUG: print 'SEARCHING FOR FULL', remainder[ iStart : iStart + 20 ]
                matchFull = self.m_patternBookExp.search( remainder, iStart )
                if not matchFull : break

                # Only skip over book because digits may belong to subsequent address.
                iStart = matchFull.end( 'book' )
                bkName = None

                # Try to decode the book (may be a bonafide book
                # we don't handle such as apocrypha in which case
                # we need to invalidate our notion of a previous
                # book so we don't try to decode the address range
                # of the apocryphal book).
                #bkNamePreviousSave = bkNamePrevious
                #bkNamePrevious = None

                bookTry = matchFull.group( 'book' )
                if DEBUG: print 'BOOKTRY', bookTry

                # Convert to canonical book name?
                bkName = self.canonicalBookName( bookTry, self.m_exactMatch )
                    
                # If found a valid bible book.
                if bkName :
                    startFull = matchFull.start( 'book' )
                    lengthFull = matchFull.end( 'book' ) - startFull
                    if DEBUG : print 'Line', lineNumber, 'MATCHFULL ', matchFull.start(), '[' + matchFull.group() + ']'
                    break

            # If we didn't see find anything looking at all like an address,
            # then we quit.  (If we found anything that could be usable--
            # even in the next scan along the string, we stay.)
            if not matchFull and not ( matchFloating and bkNamePrevious ) : break

            # We pick the nearest match to use.
            # Use the one which occurs earliest in the string.
            # (Must do this before testing for valid book name
            # or else we can't move past to look for next valid
            # match.)
            if DEBUG : print 'startFloating', startFloating
            if DEBUG : print 'startFull', startFull
            if bkNamePrevious and ( startFloating < startFull ) :
                bkName = bkNamePrevious
                matchBook = matchFloating
                matchBookStart = matchFloating.start( 'chapterandverses' )
                matchBookGroup = matchFloating.group( 'chapterandverses' )
                matchBookLength = lengthFloating
                if DEBUG : print 'Using FLOATING', '[' + matchBookGroup + ']'
            else :
                matchBook = matchFull
                matchBookStart = matchFull.start()
                matchBookGroup = matchFull.group()
                matchBookLength = lengthFull
                matchVerseSpecified = 0
                matchChapterSpecified = 0
                if DEBUG : print 'Using FULL', '[' + matchBookGroup + ']'

            # If we found a recognized book, or we found free-floating
            # verses which we assume are attached to a previous book.
            verseSeen = 0
            if DEBUG: print 'bkName', bkName, '[' + matchBookGroup + ']'
            if bkName :

                # Convert CUSTOM bookname to standard bookname.
                bookTest = bkName
                ignorecase = ( AddressFinder.MATCH_IGNORECASE == self.m_exactMatch )
                if ignorecase : bookTest = bkName.lower()
                ourBkName = None
                for index in range( len( self.m_bookNames ) ) :
                    ourBook = self.m_bookNames[ index ]
                    ourBookTest = ourBook
                    if ignorecase : ourBookTest = ourBookTest.lower()
                    if bookTest == ourBookTest :

                        # Shift index within range of our canonical names.
                        lenBooks = len( Address.BOOKNAMES )
                        iMod = index
                        while ( lenBooks <= iMod ) : iMod -= lenBooks

                        # Convert to canonical abbreviation
                        ourBkName = Address.BOOKNAMES[ iMod ]
                        break

                if not ourBkName :
                    print 'Line', lineNumber, 'error in bibletag program, invalid book recognized:', bkName
                    assert( 0 )

                # How many chapters can this book have?
                maxChapter = Address.BOOKS[ ourBkName ]
                #print ourBkName, 'can have', maxChapter, 'chapters'

                # Move past the book.
                offsetChapter = matchBookStart + matchBookLength
                chapterString = matchBookGroup[ matchBookLength : ]
                if DEBUG: print 'CHAPTERSTRING', chapterString

                # For each set of chapter and verses.
                isFirstChapter = isFirstVerse = 1
                while ( 1 ) :
                    matchChapterVerse = matchVerse = matchChapter = None

                    # May just have verses left without any chapter
                    verseString = chapterString
                    if DEBUG: print '#1 SET VERSE STRING', verseString
                    offsetVerse = 0
                    matchChapterStart = matchChapterEnd = 0
                    matchChapterVerseStart = matchChapterVerseEnd = 0
                    if DEBUG: print '#6 SET MATCH CHAPTER VERSE END', matchChapterVerseEnd
                    # Does next address indicate a semicolon or normal colon?
                    # Then we assume we are dealing with chapter, not verse.
                    nextHasColon = re.sub( '(^\D*\d+:\d+|^\s*;)', ':', chapterString )
                    nextHasColon = ( ':' == nextHasColon[ 0 ] )
                    if DEBUG: print 'NEXT HAS COLON', nextHasColon
                    if nextHasColon : isFirstVerse = 1

                    # Don't scan for a chapter if we are explicitly looking
                    # for a verse and this is the first chapter in the book
                    # without an explicit colon identifier.
                    if matchVerseSpecified and not ( nextHasColon or isFirstChapter ): 
                        if DEBUG: print 'SKIP CHAPTER SCAN'
                        break

                    # If we aren't looking explicitly for a verse and this is the
                    # first chapter or there is an explicit colon.
                    if DEBUG: print 'MATCH VERSE SPECIFIED: #3', matchVerseSpecified
                    if not matchVerseSpecified and ( nextHasColon or isFirstChapter ) :

                        if DEBUG: print 'MATCHCHAPTER CHECK', chapterString
                        matchChapterVerse = self.m_patternChapterVerseExp.search ( chapterString )
                        if ( not matchChapterVerse ) : break
                        if DEBUG: print 'MATCHCHAPTER FOUND', matchChapterVerse.group()

                        # Get the chapter.
                        matchChapter = self.m_patternChapterExp.search( matchChapterVerse.group() )
                        if ( not matchChapter ) :
                            print 'Line', lineNumber, 'BAD matchChapterVerse:', matchChapterVerse.group()
                            print 'matchBookGroup:', matchBookGroup
                            print 'chapterString:', chapterString
                            print 'original:', original
                            assert( 0 )
                        chapter = matchChapter.group( 'chapter' )

                        # If chapter is in range.
                        if int( chapter ) <= maxChapter :

                            # Prepare to scan the verse.
                            offsetVerse = matchChapterVerse.start() + matchChapter.end()
                            verseString = matchChapterVerse.group()[ matchChapter.end() : ]
                            if DEBUG: print '#2 SET VERSE STRING', verseString

                        # Else if book only has one chapter.
                        elif 1 == maxChapter :

                            # Treat chapter as a verse.
                            if DEBUG: print 'ONE CHAPTER BOOK SO TREATING CHAPTER AS VERSE', chapter
                            chapter = '1'

                        if DEBUG: print 'CHAPTER', chapter

                        # Move past what we scanned.
                        matchChapterVerseStart = matchChapterVerse.start()
                        matchChapterVerseEnd = matchChapterVerse.end()
                        if DEBUG: print '#7 SET MATCH CHAPTER VERSE END', matchChapterVerseEnd
                        matchChapterStart = matchChapter.start()
                        matchChapterEnd = matchChapter.end()


                    # Default start and end brackets the chapter on its own.
                    iStart = iStartChapter = offsetBook + offsetChapter + matchChapterVerseStart + matchChapterStart
                    iEnd   = iEndChapter = offsetBook + offsetChapter + matchChapterVerseStart + matchChapterEnd

                    # If this is the first chapter in this set of addresses, then we back up to include the book name.
                    if isFirstChapter and includeBookName :
                        iStart = offsetBook + matchBookStart

                    # For each range of verses within a single chapter.
                    while ( 1 ) :
                        lastVerse = 0

                        iVerse = offsetBook + offsetChapter + offsetVerse

                        # If there is a verse beyond the chapter.
                        if DEBUG: print 'MATCH VERSE CHECKING', verseString
                        matchVerseEnd = len( verseString )
                        matchVerse = self.m_patternVerseExp.search ( verseString )
                        if ( matchVerse ) :
                            if DEBUG: print 'MATCH VERSE FOUND', matchVerse.group(), ' FROM ', verseString
                            matchVerseEnd = matchVerse.end()
                            verse = matchVerse.group( 'verse' )

                            # We've seen a verse.
                            verseSeen = 1

                            # Calculate offsets to the decoded address.
                            iEnd = iVerse + matchVerse.end( 'verse' )
                            matchLastVerse = matchVerse.group ( 'last' )
                            if matchLastVerse : 
                                # Clean off appended alpha designators.
                                lastVerse = re.sub( '^(\d+)(\D.*)?', r'\1', matchLastVerse )
                                lastVerse = int( lastVerse )

                            # If no colon found and not an explicit verse and
                            # not a single chapter book.
                            if not nextHasColon and not matchVerseSpecified and 1 < maxChapter :

                                # If not matched by chapter scan.
                                if not matchChapter : 
                                    # Use this as the chapter.
                                    chapter = re.sub( '^(\d+)(\D.*)?', r'\1', verse )
                                    if DEBUG : print 'VERSE SCAN SETTING CHAPTER', chapter
                                    iStart = iVerse + matchVerse.start( 'verse' )
                                # Else chapter was scanned.
                                else :
                                    # Don't eat this as a verse - leave it for next chapter scan.
                                    if DEBUG : print 'VERSE SCAN NOT MOVING'
                                    matchVerseEnd = 0
                                    iEnd   = offsetBook + offsetChapter + matchChapterVerseStart + matchChapterEnd
                                    iVerse = offsetBook + offsetChapter

                                # Assume first verse within chapter.
                                verse = '1'
                                lastVerse = 0
                                isFirstVerse = 1

                                # No real verse seen yet.
                                verseSeen = 0

                                # Reset chapter match for subsequent verse scan.
                                matchChapter = None

                            # If haven't seen a verse in this chapter yet and this
                            # looks like a chapter range.
                            #DISABLED
                            if 0 and isFirstVerse and re.search( '-', matchVerse.group() ) :
                                # Assume verse one.
                                verse = '1'

                            # The offsets for subsequent verses do not include the chapter.
                            if not isFirstVerse :
                                iStart = iVerse + matchVerse.start( 'verse' )

                        # No verse found.
                        else :

                            # Only assume a missing verse for the very first try of finding one.
                            if isFirstVerse :

                                # If the book only has one chapter, assume what we have is
                                # the verse.
                                if 1 == maxChapter :
                                    verse = chapter
                                    chapter = '1'

                                # Else not a single-chapter book.
                                else :

                                    # Add the expanded address to the list with an assumed verse 1.
                                    verse = '1'
                                    if verseRange :
                                        key = ourBkName + chapter
                                        lastVerse = Address.RANGES[ key ]

                            # No verse and not first try.
                            else :

                                # Invalidate any potential chapter match.
                                matchChapterVerse = None

                        # If we recognized a chapter or verse and the verse
                        # is related to a chapter.
                        if ( matchChapterVerse or matchVerse ) and chapter :

                            # If the chapter is within range for the book.
                            key = ourBkName + chapter
                            if int( chapter ) <= maxChapter :

                                # If the verse value is within range.
                                maxVerse = Address.RANGES[ key ]
                                matchFirstVerse = re.match( '\d+', verse )
                                assert( matchFirstVerse )
                                firstVerse = matchFirstVerse.group()
                                #print 'firstVerse', firstVerse
                                #print 'lastVerse', lastVerse
                                if  int( firstVerse ) <= maxVerse and lastVerse <= maxVerse :
                                    if lastVerse : 
                                        firstVerse = firstVerse + ( '-%d' % lastVerse )
                                    #spec = ourBkName + ' ' + chapter + Address.CHAPTER_DIVISION + verse, iStart, iEnd

                                    # purge extra spaces from representation (artifacts from conversion of HTML to spaces)
                                    show = original[ iStart : iEnd ]
                                    show = re.sub( '\s+', ' ', show )
                                    #show = re.sub( '(\s+)?-(\s+)?', '-', show )
                                    show = re.sub( '(\s+)?([-,:])(\s+)?', r'\2', show )
                                    addr = Address( ourBkName, chapter, firstVerse, iStart, iEnd, show )

                                    # Reject addresses which are nested within anchors.
                                    inAnchor = 0
                                    iBeginLeft = string.rfind( lower, '<a', 0, iStart )
                                    iEndLeft = string.rfind( lower, '</a', 0, iStart )
                                    iBeginRight = string.find( lower, '<a', iStart )
                                    iEndRight = string.find( lower, '</a', iStart )
                                    if -1 != iBeginLeft :
                                        inAnchor = inAnchor or iEndLeft < iBeginLeft
                                    if -1 != iEndRight :
                                        if -1 == iBeginRight : iBeginRight = iEndRight
                                        inAnchor = inAnchor or iEndRight <= iBeginRight
                                    if inAnchor :
                                        #print iBeginLeft, iEndLeft, iStart, iBeginRight, iEndRight
                                        print 'At or before line', lineNumber, 'WARNING: skipping potential address nested in an HTML anchor:', matchBookGroup
                                    else :
                                        self.verboseLog( 'Line %s: %s' % ( lineNumber, addr ) )
                                        addresses.append( addr )
                                    if DEBUG: addr.dump()

                                    bkNamePrevious = bkName
                                    #print '5BKNAMEPREVIOUS ', bkNamePrevious
                                else :
                                    #print 'VERSE TOO LARGE', ourBkName, verse
                                    badVerse = firstVerse
                                    if  maxVerse < lastVerse : badVerse = lastVerse
                                    self.verboseLog( 'Skipping out of range verse %s in %s chapter %s : %s ' % ( badVerse, ourBkName, chapter, matchBookGroup ) )
                                
                            else :
                                self.verboseLog( 'Skipping out of range chapter %s in %s : %s' % ( chapter, ourBkName, matchBookGroup ) )

                        # Make sure we move even if no chapter or verse match.
                        if 0 == matchChapterVerseEnd : 
                            matchChapterVerseEnd = matchVerseEnd
                            if DEBUG: print '#5 SET MATCH CHAPTER VERSE END', matchChapterVerseEnd

                        # Did we find a verse?
                        if not matchVerse : break

                        # Move past the matched verse range.
                        isFirstVerse = 0
                        offsetVerse = offsetVerse + matchVerseEnd
                        verseString = verseString[ matchVerseEnd : ]
                        if DEBUG: print '#3 SET VERSE STRING', verseString

                        # If no more digits left.
                        if not re.search( '\d', verseString ) : 

                            # Advance past the verses just parsed.
                            matchChapterVerseEnd = offsetVerse

                            # Quit looking for verses.
                            break

                    # Move past the matched chapter range.
                    offsetChapter = offsetChapter + matchChapterVerseEnd
                    chapterString = chapterString[ matchChapterVerseEnd : ]
                    if DEBUG: print '#4 SET CHAPTER STRING', chapterString
                    isFirstChapter = 0
                    if not len( chapterString ) : break

            # Move to next potential book address range.
            # If failed to find a valid bible book.
            if not bkName:

                # If found what looked like a full address, but didn't match 
                # a known book. 
                if matchBook is matchFull :

                    #only skip over book name because chapter
                    # info which followed may actually be part of a subsequent
                    # valid book name (e.g., "1 Corinthians").
                    offsetBook = offsetBook + matchBook.end( 'book' )
                    remainder = remainder[ matchBook.end( 'book' ) : ]

                # Else tried to use a floating address fragment, but failed.
                else :

                    # Don't move so we include failed floating text
                    # as beginning of next full match attempt.
                    pass

            # Else found a valid bible book and address.
            else :

                # Move past entire address field.
                offsetBook = offsetBook + matchBook.end()
                remainder = remainder[ matchBook.end() : ]

                # If we saw a verse in a full address specification,
                # then subsequent floating verses should assume they are verse specifiers.
                if matchBook is matchFull and verseSeen : matchVerseSpecified = 1

# Return list of addresses, may be empty.
        if DEBUG: print '######## findAddress exit ################'
        return addresses

    def lbxBookNumber( self, bookname ) :
        if Address.BOOKNAMESLONG.count( bookname ) :
            iBook = Address.BOOKNAMESLONG.index( bookname )
            bookname = Address.BOOKNAMES[ iBook ]
        return  Address.BOOKNAME_TO_LBX_NUMBER[ bookname ]

    def canonicalBookName( self, bookTry, exactMatch = MATCH_DEFAULT ) :
        # Attempt to convert user-specified book name to a canonical name.

        # Assume not a book.
        bkName = ''

        # Only allow book names that are a minimal length.
        minLength = 2
        if bookTry and '.' == bookTry[ -1 ] : minLength = 3
        if minLength <= len( bookTry ) :

            # Special case: map certain short common abbreviations
            # which don't follow the 'closest match' rules.
            if ( 'Jn' == bookTry[ 0 : 2 ] ) : bookTry = 'John'

            # Does case matter?
            if AddressFinder.MATCH_IGNORECASE == exactMatch :
                bookTry = bookTry.lower()

            # Convert special prefixes?
            if AddressFinder.MATCH_EXACT != self.m_exactMatch: 
                # Must be applied from longest pattern to shortest order.
                bookTry = self.m_patternThirdExp.sub( '3', bookTry )
                bookTry = self.m_patternSecondExp.sub( '2', bookTry )
                bookTry = self.m_patternFirstExp.sub( '1', bookTry )

            # Get rid of redundant spaces in input pattern.
            bookTry = re.sub( '\s+', ' ', bookTry )

            # For each book we know about.
            fewestMissed = 100
            bestBook = None
            for book in self.m_bookNames :
                bookPat = book

                # Does case matter?
                if AddressFinder.MATCH_IGNORECASE == self.m_exactMatch :
                    bookPat = bookPat.lower()

                # Assume no match.
                iBook = -1

                # Check for exact match?
                if AddressFinder.MATCH_EXACT == self.m_exactMatch: 
                    iBook = string.find( bookTry, bookPat )
                    if ( iBook == 0 ) :
                        bestBook = bookPat
                        break;

                # Else allow abbreviations which are strict subsets
                # but they must be at least 2 characters long.
                else :

                    # Get rid of final period on abbreviated trial
                    # book.
                    if '.' == bookTry[ -1 ] : bookTry = bookTry[ : -1 ]

                    # For each letter in the trial book name.
                    iPat = 0
                    iTry = 0
                    missed = 0
                    while ( iPat < len( bookPat ) and iTry < len( bookTry ) ) :

                        # Get corresponding letters.
                        letterTry = bookTry[ iTry ]
                        letterPat = bookPat[ iPat ]

                        # If letters match.
                        if letterTry == letterPat : 

                            # Move to next position.
                            iPat = iPat + 1
                            iTry = iTry + 1

                            # If input is exhausted, we are done with this one.
                            if len( bookTry ) <= iTry : 

                                # Keep track of best.
                                if missed < fewestMissed :
                                    fewestMissed = missed
                                    #print 'fewestMissed', fewestMissed, bookPat
                                    bestBook = bookPat
                                break

                        # Else letters don't match.
                        else :

                            # First letter MUST match.
                            if iTry == 0 : break

                            # Move to next in pattern.
                            # Can't advance trial book pattern.
                            iPat = iPat + 1

                            # Keep track of missed matches for
                            # best match
                            missed = missed + 1

                    # Quit early if we find an exact substring.
                    if 0 == fewestMissed : 
                        break

            if ( bestBook ) : bkName = bestBook

        return bkName

    def verboseLog( self, text ) :
        if self.m_verbose :
            print text
            if self.m_verboseFile : self.m_verboseFile.write( text + '\n' )

